/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E11 (compare with E3).

      E ::= E '+' T    // left recursion means left associative
          | E '-' T    // left recursion means left associative
          | T

      T ::= T '*' F    // left recursion means left associative
          | T '/' F    // left recursion means left associative
          | F

      F ::= '-' F      // right recursion means right associative
          | '+' F      // right recursion means right associative
          | '(' E ')'
          | 'a'

   We can rewrite the grammar this way.

      E ::= T ( ( '+' | '-' ) T )*   // sum of Terms (iteration)
      T ::= F ( ( '*' | '/' ) F )*   // product of Factors (iteration)
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | 'a'

   This parser builds a parse tree for the
   expression being parsed. It will throw an
   error if the string cannot be parsed.

   Some example strings are:
      "-a"
      "+a"
      "+(-a)"
      "-(+a)"
      "+-(-+a)"
      "(+a + -a)"
      "-(a + a) / -a"
      "-a * -(a + a)"
      "(a + a) * -(a / a)"
      "-a / (a + -(a * -a))"
      "-(-(-a + -a) * -a) + -a"
   This grammar can handle any valid four function
   arithmetic expression.
*/
public class Parse2Syntax_E11v2
{
   public static Tree<String> parseE11(String stringInE11)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE11);
      // Parse the list of tokens.
      final Tree<String> parseTree = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return parseTree;
   }


   /**
      Left recursion,
         E ::= E [ ( '+' | '-' ) T ]
      converted to iteration,
         E ::= T ( ( '+' | '-' ) T )*
   */
   private static Tree<String> getE(Tokenizer tokens)
   {
      Tree<String> parse = new LinkedTree<>("E", getT(tokens));
      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("+")
                                   || tokens.peek().lexeme.equals("-") ) )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<String> subTree = getT(tokens); // recursion
         parse = new LinkedTree<>("E",
                                  parse,
                                  new LinkedTree<>(tk.toString()),
                                  subTree);
      }
      return parse;
   }


   /**
      Left recursion,
         T ::= T [ ( '*' | '/' ) F ]
      converted to iteration,
         T ::= F ( ( '*' | '/' ) F )*
   */
   private static Tree<String> getT(Tokenizer tokens)
   {
      Tree<String> parse = new LinkedTree<>("T", getF(tokens));
      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("*")
                                   || tokens.peek().lexeme.equals("/") ) )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<String> subTree = getF(tokens); // recursion
         parse = new LinkedTree<>("T",
                                  parse,
                                  new LinkedTree<>(tk.toString()),
                                  subTree);
      }
      return parse;
   }


   /**
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | 'a'
   */
   private static Tree<String> getF(Tokenizer tokens)
   {
      final Tree<String> parse;
      if ( tokens.peek().lexeme.equals("-")
        || tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<String> subTree = getF(tokens);
         parse = new LinkedTree<>("F",
                                  new LinkedTree<>(tk.toString()),
                                  subTree);
      }
      else if ( tokens.peek().lexeme.equals("(") )
      {
         final Token tk1 = tokens.next(); // consume the "("
         final Tree<String> subTree = getE(tokens);
         if ( ! tokens.peek().lexeme.equals(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
         final Token tk2 = tokens.next(); // consume the ")"
         parse = new LinkedTree<>("F",
                                  new LinkedTree<>(tk1.toString()),
                                  subTree,
                                  new LinkedTree<>(tk2.toString()));
      }
      else if ( tokens.peek().lexeme.equals("a") )
      {
         final Token tk = tokens.next(); // consume the "a"
         parse =  new LinkedTree<>("F",
                                   new LinkedTree<>(tk.toString()));  // consume the "a"
      }
      else
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      return parse;
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "-a",
                              "+a",
                              "+(-a)",
                              "-(+a)",
                              "+-(-+a)",
                              "(+a + -a)",
                              "-(a + a) / -a",
                              "-a * -(a + a)",
                              "(a + a) * -(a / a)",
                              "-a / (a + -(a * -a))",
                              "-(-(-a + -a) * -a) + -a",
                              "-a * -(-a + -a) / +a",
                              "-a / -(+(-a - -a)) * -a",
                              "a - (a * a) + a",
                              "a * (a + a) - a",
                              "a * (a + a) - -a",
                              "-(a + a + a) - a",
                              "a + -(a * a)) + a",
                              "a + -(-(a * a) + a"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<String> parseTree = parseE11( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "Parse Tree = " + parseTree + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E11v2_Exp_%02d_Parse", i);
            Tree2dot.tree2dot(parseTree, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
