/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E3.

      E ::= E '+' 'a'   // left recursion means left associative
          | 'a'

   We cannot parse this language, as it is written, using a
   recursive descent parser. The left recursion in the grammar
   causes infinite recursion in the recursive descent parser.
   We must rewrite the grammar to remove the left recursion
   and replace the recursion with a kind of iteration.

   The above grammar can be rewritten to use the Kleene star
   (from "regular expressions") in place of the left recursion.
   The new grammar needs only one production.

      E ::= 'a' ( '+' 'a' )*  // * means zero-or-more

   This parser builds a parse tree for the expression
   being parsed. Even though this parser is parsing the
   grammar with the Kleene star, this parser builds the
   parse tree fro the grammar with left recursion. It
   will throw an error if the string cannot be parsed.

   Some example strings are:
      "a"
      "a + a"
      "a + a + a"
      "a + a + a + a"
*/
public class Parse2Syntax_E3
{
   public static Tree<String> parseE3(String stringInE3)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE3);
      // Parse the list of tokens.
      final Tree<String> parseTree = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return parseTree;
   }


   /**
      E ::= 'a' ( '+' 'a' )*
   */
   private static Tree<String> getE(Tokenizer tokens)
   {
      final Token tk = tokens.next();
      if ( ! tk.lexeme.equals("a") )
      {
         throw new ParseError("expected \"a\", found " + tk);
      }
      Tree<String> parse = new LinkedTree<>("E",
                                            new LinkedTree<>(tk.toString()));

      while ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk1 = tokens.next(); // consume the "+"
         if ( ! tokens.peek().lexeme.equals("a") )
         {
            throw new ParseError("expected \"a\", found " + tokens.peek());
         }
         final Token tk2 = tokens.next(); // consume the "a"
         parse = new LinkedTree<>("E",
                                  parse,
                                  new LinkedTree<>(tk1.toString()),
                                  new LinkedTree<>(tk2.toString()));
      }
      return parse;
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a",
                              "a + a",
                              "a + a + a",
                              "a + a + a + a",
                              "a+a+a+a",
                              "a+a+a-a",
                              "a + a +",
                              "+ a + a"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<String> parseTree = parseE3( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "Parse Tree = " + parseTree + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E3_Exp_%02d_Parse", i);
            Tree2dot.tree2dot(parseTree, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
