/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E6 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= '(' E ')'
          | 'a'

   We can rewrite the grammar this way.

      E ::= T [ '+' E ]
      T ::= '(' E ')'
          | 'a'

   This parser builds a parse tree for the
   expression being parsed. It will throw an
   error if the string cannot be parsed.

   Some example strings are:
      "(a + a)"
      "(a + a) + a"
      "a + (a + a)"
      "(a + a) + (a + a)"
      "a + (a + (a + a))"
      "((a + a) + a) + a"
      "a + (a + a) + a"
*/
public class Parse2Syntax_E6
{
   public static Tree<String> parseE6(String stringInE6)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE6);
      // Parse the list of tokens.
      final Tree<String> ast = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return ast;
   }


   /**
      E ::= T [ '+' E ]
   */
   private static Tree<String> getE(Tokenizer tokens)
   {
      final Tree<String> parse = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next();  // consume the "+"
         final Tree<String> subTree = getE(tokens); // recursion
         return new LinkedTree<>("E",
                                 parse,
                                 new LinkedTree<>(tk.toString()),
                                 subTree);
      }
      else
      {
         return new LinkedTree<>("E", parse);
      }
   }


   /**
      T ::= '(' E ')'
          | 'a'
   */
   private static Tree<String> getT(Tokenizer tokens)
   {
      final Tree<String> parse;
      if ( tokens.peek().lexeme.equals("(") )
      {
         final Token tk1 = tokens.next(); // consume the "("
         final Tree<String> subTree = getE(tokens);
         if ( ! tokens.peek().lexeme.equals(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
         final Token tk2 = tokens.next(); // consume the ")"
         parse = new LinkedTree<>("T",
                                  new LinkedTree<>(tk1.toString()),
                                  subTree,
                                  new LinkedTree<>(tk2.toString()));
      }
      else if ( tokens.peek().lexeme.equals("a") )
      {
         final Token tk = tokens.next(); // consume the "a"
         parse =  new LinkedTree<>("T",
                                   new LinkedTree<>(tk.toString())); // consume the "a"
      }
      else
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      return parse;
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "(a + a)",
                              "(a + a) + a",
                              "a + (a + a)",
                              "(a + a) + (a + a)",
                              "a + (a + (a + a))",
                              "((a + a) + a) + a",
                              "a + (a + a) + a",
                              "a + ((a + a)) + a",
                              "a - (a + a) + a",
                              "a - (a + a) - a",
                              "a + (a + a)) + a",
                              "a + ((a + a) + a",
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<String> parseTree = parseE6( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "Parse Tree = " + parseTree + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E6_Exp_%02d_Parse", i);
            Tree2dot.tree2dot(parseTree, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
