/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E7 (with right recursion).

      E ::= 'a' '+' E   // right recursion means right associative
          | 'a' '-' E   // right recursion means right associative
          | 'a'

   We can rewrite the grammar this way.

      E ::= 'a' [ ( '+' | '-' ) E ]

   This production says that an E expands into an
   'a' followed by an optional plus-or-minus E

   This parser builds a parse tree for the
   expression being parsed. It will throw an
   error if the string cannot be parsed.

   Some example strings are:
      "a"
      "a + a"
      "a - a"
      "a - a + a"
      "a + a - a - a"
*/
public class Parse2Syntax_E7
{
   public static Tree<String> parseE7(String stringInE7)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE7);
      // Parse the list of tokens.
      final Tree<String> parseTree = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return parseTree;
   }


   /**
      E ::= 'a' [ ( '+' | '-' ) E ]
   */
   private static Tree<String> getE(Tokenizer tokens)
   {
      final Token tk1 = tokens.next();
      if ( ! tk1.lexeme.equals("a") )
      {
         throw new ParseError("expected \"a\", found " + tk1);
      }
      if ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("+")
                                || tokens.peek().lexeme.equals("-") ) )
      {
         final Token tk2 = tokens.next(); // consume the operator
         final Tree<String> parse = getE(tokens); // recursion
         return new LinkedTree<>("E",
                                 new LinkedTree<>(tk1.toString()),
                                 new LinkedTree<>(tk2.toString()),
                                 parse);
      }
      else
      {
         return new LinkedTree<>("E",
                                 new LinkedTree<>(tk1.toString()));
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a",
                              "a + a",
                              "a - a",
                              "a - a + a",
                              "a + a - a - a",
                              "a-a-a+a",
                              "a+a+a-a",
                              "a + + a + a",
                              "a + a a + a",
                              "a - a +",
                              "- a + a",
                              "+ a - a"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<String> parseTree = parseE7( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "Parse Tree = " + parseTree + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E7_Exp_%02d_Parse", i);
            Tree2dot.tree2dot(parseTree, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
