/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E8 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= 'a' '*' T  // right recursion means right associative
          | 'a'

   We can rewrite the grammar this way.

      E ::=  T  [ '+' E ]
      T ::= 'a' [ '*' T ]

   This parser builds a parse tree for the
   expression being parsed. It will throw an
   error if the string cannot be parsed.

   Some example strings are:
      "a + a"
      "a * a"
      "a + a * a"
      "a * a + a"
      "a + a * a + a"
      "a * a + a * a"
      "a * a * a * a"
      "a * a * a + a * a"
*/
public class Parse2Syntax_E8
{
   public static Tree<String> parseE8(String stringInE8)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE8);
      // Parse the list of tokens.
      final Tree<String> parseTree = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return parseTree;
   }


   /**
      E ::=  T  [ '+' E ]
   */
   private static Tree<String> getE(Tokenizer tokens)
   {
      final Tree<String> parse = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next(); // consume the "+"
         final Tree<String> subTree = getE(tokens); // recursion
         return new LinkedTree<>("E",
                                 parse,
                                 new LinkedTree<>(tk.toString()),
                                 subTree);
      }
      else
      {
         return new LinkedTree<>("E", parse);
      }
   }


   /**
      T ::= 'a' [ '*' T ]
   */
   private static Tree<String> getT(Tokenizer tokens)
   {
      if ( ! tokens.peek().lexeme.equals("a") )
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      final Token tk1 = tokens.next(); // consume the "a"
      final Tree<String> parse = new LinkedTree<>(tk1.toString());
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("*") )
      {
         final Token tk2 = tokens.next(); // consume the "*"
         final Tree<String> subTree = getT(tokens); // recursion
         return new LinkedTree<>("T",
                                 parse,
                                 new LinkedTree<>(tk2.toString()),
                                 subTree);

      }
      else
      {
         return new LinkedTree<>("T", parse);
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a + a",
                              "a * a",
                              "a + a * a",
                              "a * a + a",
                              "a + a * a + a",
                              "a * a + a * a",
                              "a * a * a * a",
                              "a * a * a + a * a",
                              "a * + a",
                              "+ a * a",
                              "* a * a",
                              "a + + a",
                              "a * * a",
                              "a * a * ",
                              "a + a * ",
                              "a * a + "
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<String> parseTree = parseE8( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "Parse Tree = " + parseTree + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E8_Exp_%02d_Parse", i);
            Tree2dot.tree2dot(parseTree, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
