/*

*/

package tree;

import java.util.List;

/**
   Define an interface for trees and then
   use that interface to implement a number
   of methods that act on trees.
*/
public interface Tree<T>
{
   /**
      @return a reference to the item in the root node of this tree
   */
   public T root();

   /**
      @return a reference to the {@link List} of sub trees to this {@code Tree}
   */
   public List<Tree<T>> subTrees();


   // Use the above two interface methods to define
   // all the recursive methods we need on trees.
   // All the following methods are "default" (that
   // is, concrete) methods of this interface.

   /**
      Determine if a tree is a leaf node.

      @return true if this {@code Tree} is just a single node
   */
   public default boolean isLeaf()
   {
      return subTrees().isEmpty();
   }


   /**
      Compute the maximum degree of a {@code Tree}.

      @return the maximum degree of any node in this {@code Tree}
   */
   public default int degree()
   {
      if ( isLeaf() )
      {
         return 0;
      }
      else // an internal node
      {
         int result = subTrees().size();

         for ( final Tree<T> tree : subTrees() )
         {
            final int degree = tree.degree();  // recursion
            result = Math.max(result, degree);
         }

         return result;
      }
   }


   /**
      Calculate the height of a {@code Tree}.

      @return the height of this {@code Tree}
   */
   public default int height()
   {
      if ( isLeaf() )
      {
         return 0;
      }
      else // an internal node
      {
         // find the highest sub tree
         int result = 0;
         for ( final Tree<T> tree : subTrees() )
         {
            final int height = tree.height();  // recursion
            result = Math.max(result, height);
         }
         return 1 + result;
      }
   }


   /**
      Calculate the size of a {@code Tree}.

      @return the number of leaves and internal nodes in this {@code Tree}
   */
   public default int size()
   {
      if ( isLeaf() )
      {
         return 1;
      }
      else // an internal node
      {
         // sum up the sizes of sub trees
         int result = 0;
         for ( final Tree<T> tree : subTrees() )
         {
            result += tree.size(); // recursion
         }
         return 1 + result;
      }
   }


   /**
      Do a pre-order, depth-first-traversal of a {@code Tree}.

      @return a {@link String} representation of the pre-order depth-first traversal of this {@code Tree}
   */
   public default String preOrder()
   {
      if ( isLeaf() )
      {
         return root() + " ";
      }
      else // an internal node
      {
         String result = "";
         result += root() + " ";
         for ( final Tree<T> tree : subTrees() )
         {
            result += tree.preOrder();  // recursion
         }
         return result;
      }
   }


   /**
      Do a post-order, depth-first-traversal of a {@code Tree}.

      @return a {@link String} representation of the post-order depth-first traversal of this {@code Tree}
   */
   public default String postOrder()
   {
      if ( isLeaf() )
      {
         return root() + " ";
      }
      else // an internal node
      {
         String result = "";
         for ( final Tree<T> tree : subTrees() )
         {
            result += tree.postOrder();  // recursion
         }
         result += root() + " ";
         return result;
      }
   }


   /**
      Force any implementation of {@code Tree} to
      convert a {@code Tree} into a {@link String}.

      @return a {@link String} representation of the {@code Tree}
   */
   @Override
   public String toString();
}
